Tested x86_64-pc-linux-gnu, applying to trunk.

-- >8 --

An experiment led me to notice that in some cases we were ending up with
TARGET_EXPR initialized by TARGET_EXPR, which isn't useful.

The target_expr_needs_replace change won't make a difference in most cases,
since cp_genericize_init will have already expanded VEC_INIT_EXPR by the
time we consider it, but it is correct.

gcc/cp/ChangeLog:

        * cp-gimplify.cc (cp_fold_r) [TARGET_EXPR]: Collapse
        TARGET_EXPR within TARGET_EXPR.
        * constexpr.cc (cxx_eval_outermost_constant_expr): Avoid
        adding redundant TARGET_EXPR.
        * cp-tree.h (target_expr_needs_replace): VEC_INIT_EXPR doesn't.
---
 gcc/cp/cp-tree.h      |  3 ++-
 gcc/cp/constexpr.cc   |  4 ++--
 gcc/cp/cp-gimplify.cc | 16 +++++++++++++---
 3 files changed, 17 insertions(+), 6 deletions(-)

diff --git a/gcc/cp/cp-tree.h b/gcc/cp/cp-tree.h
index 3cbcdf726ca..d696fd54a7a 100644
--- a/gcc/cp/cp-tree.h
+++ b/gcc/cp/cp-tree.h
@@ -5426,7 +5426,8 @@ target_expr_needs_replace (tree t)
     return false;
   while (TREE_CODE (init) == COMPOUND_EXPR)
     init = TREE_OPERAND (init, 1);
-  return TREE_CODE (init) != AGGR_INIT_EXPR;
+  return (TREE_CODE (init) != AGGR_INIT_EXPR
+         && TREE_CODE (init) != VEC_INIT_EXPR);
 }
 
 /* True if EXPR expresses direct-initialization of a TYPE.  */
diff --git a/gcc/cp/constexpr.cc b/gcc/cp/constexpr.cc
index ed41d755269..db7571d7d71 100644
--- a/gcc/cp/constexpr.cc
+++ b/gcc/cp/constexpr.cc
@@ -8065,8 +8065,8 @@ cxx_eval_outermost_constant_expr (tree t, bool 
allow_non_constant,
       if (TREE_CODE (t) == TARGET_EXPR
          && TARGET_EXPR_INITIAL (t) == r)
        return t;
-      else if (TREE_CODE (t) == CONSTRUCTOR)
-       ;
+      else if (TREE_CODE (t) == CONSTRUCTOR || TREE_CODE (t) == CALL_EXPR)
+       /* Don't add a TARGET_EXPR if our argument didn't have one.  */;
       else if (TREE_CODE (t) == TARGET_EXPR && TARGET_EXPR_CLEANUP (t))
        r = get_target_expr (r);
       else
diff --git a/gcc/cp/cp-gimplify.cc b/gcc/cp/cp-gimplify.cc
index c05be833357..73548888783 100644
--- a/gcc/cp/cp-gimplify.cc
+++ b/gcc/cp/cp-gimplify.cc
@@ -1084,9 +1084,9 @@ cp_fold_r (tree *stmt_p, int *walk_subtrees, void *data_)
        }
       break;
 
-      /* These are only for genericize time; they're here rather than in
-        cp_genericize to avoid problems with the invisible reference
-        transition.  */
+      /* cp_genericize_{init,target}_expr are only for genericize time; they're
+        here rather than in cp_genericize to avoid problems with the invisible
+        reference transition.  */
     case INIT_EXPR:
       if (data->genericize)
        cp_genericize_init_expr (stmt_p);
@@ -1095,6 +1095,16 @@ cp_fold_r (tree *stmt_p, int *walk_subtrees, void *data_)
     case TARGET_EXPR:
       if (data->genericize)
        cp_genericize_target_expr (stmt_p);
+
+      /* Folding might replace e.g. a COND_EXPR with a TARGET_EXPR; in
+        that case, use it in place of this one.  */
+      if (tree &init = TARGET_EXPR_INITIAL (stmt))
+       {
+         cp_walk_tree (&init, cp_fold_r, data, NULL);
+         *walk_subtrees = 0;
+         if (TREE_CODE (init) == TARGET_EXPR)
+           *stmt_p = init;
+       }
       break;
 
     default:

base-commit: bbdcdf5cc73e1b3385d9a25bdab4df70c4bd8c2e
-- 
2.31.1

Reply via email to